/* drvVmi5588.c -  Device driver for VMIC VMIVME5588 */
/* $Id: drvVmi5588.c,v 1.4 2002/01/18 01:27:28 cboyer Exp $
*
*   Author:    Andrew Johnson
*   Date:      10-10-94
*
* Experimental Physics and Industrial Control System (EPICS)
*
* Developed at the Royal Greenwich Observatory for the Gemini
* 8M Telescopes Project.
*
* This driver incorporates some code from Bob Daly's VmiVme5576
* driver drvRM.c, dated 23 Oct 93, but is mostly new.
*/

/*
modification history
--------------------


DESCRIPTION
This driver contains support routines for the VMIC VMIVME5588 Reflective
Memory card, for use both from EPICS and from a vxWorks/C subroutine.

INCLUDE FILES: vmi5588.h
*/

/* vxWorks #includes */
#include   <vxWorks.h>
#include   <vme.h>
#include   <rebootLib.h>
#include   <iv.h>
#include   <intLib.h>
#include   <symLib.h>
#include   <vxLib.h>
#include   <logLib.h>
#include   <stdio.h>
#include   <sysLib.h>
#include   <string.h>

#ifndef NO_EPICS
    /* EPICS #includes */
    #include       <dbDefs.h>
    #include       <dbScan.h>
    #include       <drvSup.h>
    #include       <devSup.h>
    #include       <devLib.h>
    #include       <module_types.h>
    #include       <task_params.h>
#endif
#include   "vmi5588.h"

/* Board addressing */

#define RM_VME_BASE   0xA00000

#define RM_VME_SPACE  VME_AM_STD_SUP_DATA

/* Interrupt addressing */
#define RM_INT_VECTOR   0xb0  /* 4 vectors used, b0 to b3 */
#define RM_INT_LEVEL    6     /* interrupts are urgent */

/* Internal driver sizes */
#define RM_PAGE_SIZE 0x0400   /* 1024 bytes per page */
#define RM_NUM_PAGE  255      /* max number of pages */

#define RM_SYM_HASHSIZE    6  /* Hash size for symbol table (log2) */
#define RM_INPUTLINESIZE 256  /* Input line buffer length */
#define RM_INPUTTYPESIZE  20  /* type keyword buffer length */
#define RM_INPUTNAMESIZE 128  /* name buffer length */

/* Storage to allocate for each shared rm_data structure */
#define RM_SIZE_ALOG 16
#define RM_SIZE_LONG 12
#define RM_SIZE_STRG 48
#define RM_SIZE_ARRY 16       /* Plus size of array */

/* Board identification */
#define RM_BOARD_5578      0x27
#define RM_BOARD_5588      0x42

/* intRxStatus bits */
#define RM_IRS_INT1        0x01
#define RM_IRS_INT2        0x02
#define RM_IRS_INT3        0x04
#define RM_IRS_RX_SIG      0x10
#define RM_IRS_VIOLATION   0x20
#define RM_IRS_LATCHED     0x40
#define RM_IRS_RX_SYNC     0x80

/* boardCsr bits */
#define RM_CSR_FAST        0x01
#define RM_CSR_MASK        0x02
#define RM_CSR_OWN_DATA    0x04
#define RM_CSR_BAD_DATA    0x08
#define RM_CSR_TX_EMPTY    0x10
#define RM_CSR_TX_HALF     0x20
#define RM_CSR_RX_HALF     0x40
#define RM_CSR_FAIL        0x80

/* cmd_Node bits */
#define RM_CMD_RESET       0x0000
#define RM_CMD_INT1        0x0100
#define RM_CMD_INT2        0x0200
#define RM_CMD_INT3        0x0300
#define RM_CMD_BROADCAST   0x4000

/* interrupt[n].control bits */
#define RM_CR_INT_AUTOCLR   0x08
#define RM_CR_INT_ENABLE    0x10

/* Local forward reference */
LOCAL void      vmi5588_reboot(int startType);


/* Symbol Type data */
static struct {
    char           *ptokString;
    unsigned char   tokType;
} rmToken[] = {
    {
   "page", RM_TYPE_PAGE
    },
    {
   "analogue", RM_TYPE_ALOG
    },
    {
   "long", RM_TYPE_LONG
    },
    {
   "string", RM_TYPE_STRG
    },
    {
   "array", RM_TYPE_ARRY
    },
    {
   "user", RM_TYPE_USER
    },
    {
   NULL, 0
    }
};


/* Driver variables */
LOCAL VOIDFUNCPTR pisr[4];
LOCAL SYMTAB_ID rmSymTbl;
int rmMaxAttempts;
int vmi5588Debug;


/* Stuff used by the EPICS driver only */
#ifndef NO_EPICS
struct {
    long            number;
    struct {
   short           page;
   short           lastPageFlag;
   IOSCANPVT       ioscanpvt;
    } p[RM_NUM_PAGE];
} pageIo;

/* Driver support DRVET */
struct {
    long            number;
    DRVSUPFUN       report;
    DRVSUPFUN       init;
} drvVmi5588 = {
    2,
    vmi5588_report,
    vmi5588_init
};

#endif


/* Memory layout of the reflective memory board */
struct {                     /* VMIVME5588    */
    char                     pad1;                   /* unused        */
    char                     boardId;                /* BID           */
    volatile unsigned char   intRxStatus;            /* IRS           */
    char                     pad2;                   /* unused        */
    unsigned char            nodeId;                 /* NID           */
    volatile unsigned char   boardCsr;               /* CSR           */
    volatile unsigned short  cmd_Node;               /* CMD & CMDN    */
    char                     pad3[24];               /* unused        */
    struct {
   char                      pad4[2];                /* unused        */
   volatile unsigned char    senderId;               /* SIDn          */
   volatile unsigned char    control;                /* CRn           */
    }                        interrupt[4];           /* for n=0 to 3  */
    struct {
   char                      pad5[3];                /* unused        */
   volatile unsigned char    number;                 /* VRn           */
    }                        vector[4];              /* for n=0 to 3  */
    /* Below here the card is just reflected RAM */
    volatile int             test;                   /* to check ring */
    char                     pad6[0x100-0x44];       /* align to xx00 */
    volatile short           pageFlag[RM_NUM_PAGE];  /* Update flags  */
    short                    pad7[256-RM_NUM_PAGE];  /* align to xx00 */
    char                     pad8[0x100];            /* align to x400 */
    unsigned char   mem[RM_NUM_PAGE*RM_PAGE_SIZE];   /* data storage  */
} *prm;

/*************************************************************************
*
* vmi5588_init - DRVET Init function
*
* This routine initialises the vmi5588 driver.  It is called by EPICS in
* iocInit() as part of the driver initialisation process.
*
* It registers the card with the devLib address routines, and makes sure
* that a VMIC5588 card is present at the given location. 
*
* RETURNS:
* OK, or S_dev_??? errors.
*/

long vmi5588_init
(
    void
)
{
    int             i;
    long            status;
    char            test;

    /* Work out what the board is */
    status = sysBusToLocalAdrs(RM_VME_SPACE, (char *) RM_VME_BASE, 
                               (char **) &prm);
    if (status != OK)
       return status;
    
    /* Check if something's out there */
    if (vxMemProbe(&prm->boardId, READ, sizeof(char), &test) != OK) {
       /* No there isn't, but don't complain about it here */
       prm = NULL;
    } else {
       /* is this the right card? */
       if (prm->boardId != RM_BOARD_5578 && prm->boardId != RM_BOARD_5588) {
           prm = NULL;
           return S_dev_wrongDevice;
       }

       if (vmi5588Debug) {
           printf("vmi5588_init: Found RM card at addr %x\n", (int) prm);
       }

       /* enable VMEbus interrupts onto the CPU card */
       status = sysIntEnable(RM_INT_LEVEL);
       if (status != OK)
          return status;

       /* initialise the card interrupters */
       for (i = 0; i <= 3; i++) {
           pisr[i] = NULL;
           prm->vector[i].number = RM_INT_VECTOR + i;
       }

       /* turn any interrupts off if we do a reboot */
       rebootHookAdd((FUNCPTR) vmi5588_reboot);

       rmMaxAttempts = 0;

       /* Finally we turn off the FAIL LED */
       prm->boardCsr &= ~RM_CSR_FAIL;
    }
    return OK;
}

/*****************************************************************************
*
* vmi5588_report - DRVET Report function
*
* This is the driver status report function, which displays the current
* status of the hardware when called, usually from dbior().
*
* RETURNS:
* OK, or S_dev_NoInit if driver not initialised.
*
* EXAMPLE:
* .CS
*   -> vmi5588_report
*   vmi5588: RM node 0x1, status 0x40, max 0 retries
*   value = 0 = 0x0
* .CE
*/

long vmi5588_report
(
    void
)
{
    if (prm == NULL)
       return S_dev_NoInit;

    printf("vmi5588: RM node 0x%x, status 0x%x, max %d retries\n",
           prm->nodeId, (int)(rmStatus(0L)), rmMaxAttempts);

    if (vmi5588Debug) {
       unsigned char   irs, csr, icr;
       int             i;

       printf("test address = 0x%x, mem starts at 0x%x\n", 
              (int) &prm->test, (int) prm->mem);
       
       /* read status */
       irs = prm->intRxStatus;
       csr = prm->boardCsr;

       printf("   Receiver: %s, PLL %s%s%s\n",
              irs & RM_IRS_RX_SIG ? "NO INPUT SIGNAL" : "Input signal good",
              irs & RM_IRS_VIOLATION ? "RESYNC NEEDED" : "Locked",
              irs & RM_IRS_LATCHED ? ", Recent Sync loss" : "",
              irs & RM_IRS_RX_SYNC ? ", SYNC BIT HIGH" : "");

       printf("   Jumpers:  %s mode, Transfer Error Interrupt %s\n",
              csr & RM_CSR_FAST ? "Fast" : "Slow",
              csr & RM_CSR_MASK ? "Disabled" : "Enabled");

       printf("   Status:   Fibre ring %s, %sFail LED %s\n",
              csr & RM_CSR_OWN_DATA ? "Intact" : "BROKEN",
              csr & RM_CSR_BAD_DATA ? "TRANSFER ERROR, " : "",
              csr & RM_CSR_FAIL ? "ON" : "Off");

       printf("   FIFOs:    Transmitter %s, Receiver %s\n",
              csr & RM_CSR_TX_EMPTY ? csr & RM_CSR_TX_HALF ?
              "<50% Full" : ">50% FULL" : "Empty",
              csr & RM_CSR_RX_HALF ? "<50% Full" : ">50% FULL");

       printf("   Int's:    %s%s%s\n",
              irs & RM_IRS_INT1 ? "Irq 1 pending " : "",
              irs & RM_IRS_INT2 ? "Irq 2 pending " : "",
              irs & RM_IRS_INT3 ? "Irq 3 pending " : "");

       for (i = 0; i <= 3; i++) {
           icr = prm->interrupt[i].control;

           printf("      Int %d: %s, Level %d %s, %svector %x\n", i,
                  pisr[i] != 0 ? "Allocated" : "Not in use",
                  icr & 7,
                  icr & RM_CR_INT_ENABLE ? "enabled" : "disabled",
                  icr & RM_CR_INT_AUTOCLR ? "Auto clear, " : "",
                  prm->vector[i].number);
       }
    }

    return OK;
}

/*****************************************************************************
*
* vmi5588_reboot - rebootHook routine
*
* This routine is connected to the vxWorks reboot Hook by vmi5588_init,
* and is responsible for turning off RM interrupts. This is necessary
* because other nodes may continue to send interrupts after we have
* restarted.
*
* RETURNS: N/A
*
* NOMANUAL
*/

LOCAL void vmi5588_reboot
(
    int startType      /* vxWorks reboot type */
)
{
    int             i;

    for (i = 0; i <= 3; i++)
        prm->interrupt[i].control &= ~RM_CR_INT_ENABLE;
}


/*****************************************************************************
*
* vmi5588_intr - interrupt handler
*
* All RM interrupts pass through this routine, which reads the RM source ID
* and calls the connected routine for this channel.
*
* RETURNS: N/A
*
* NOMANUAL
*/

void vmi5588_intr
(
    int irqNumber      /* RM interrupt channel number */
)
{
    if (prm == NULL || irqNumber < 0 || irqNumber > 3) {
       logMsg("vmi5588_intr: Bad RM Interrupt, parameter = 0x%x", 
              irqNumber, 0,0,0,0,0);
       return;
    }

    if (pisr[irqNumber] != NULL) {
       if (irqNumber > 0)
          (*pisr[irqNumber]) (prm->interrupt[irqNumber].senderId);
    else
       (*pisr[irqNumber]) (0);

    /* finally re-initialise the interrupt hardware */
    prm->interrupt[irqNumber].control = RM_INT_LEVEL | 
                                        RM_CR_INT_ENABLE | 
                                        RM_CR_INT_AUTOCLR;
    }
    else
       logMsg("vmi5588_intr: RM Interrupt #%d received while disconnected", 
              irqNumber, 0,0,0,0,0);
}

/*****************************************************************************
*
* rmIntConnect - Connect a C routine up to an RM Interrupt
*
* This is used to register an interrupt routine with the RM driver layer,
* and initialises the BIM registers on the 5588 card to enable interrupts
* for the given RM interrupt channel number.
*
* The interrupt routine should expect a single integer parameter, which 
* will contain the node number of the RM interrupt source.
*
* NOTE
* Channel 0 is sent by the 5588 card to indicate transfer errors, or that
* the transmit FIFO is more than half-full. This interrupt is not used by
* the EPICS driver.
* .br
* Channel 1 is used by the device support layer for EPICS RM record types.
* .br
* Channels 2 and 3 are available for general C subroutines.
*
* RETURNS:
* OK, or S_dev_??? errors.
* 
* EXAMPLE:
* .CS
*   void myIsr(int nodeFrom);
*   int status;
*
*   status = rmIntConnect(2, myIsr);
*   if(status != 0)
*       printf("Can't connect RM interrupt: %d\\n", status);
* .CE
*
* SEE ALSO: rmIntDisconnect(), rmIntSend()
*/

long rmIntConnect
(
    int irqNumber,         /* RM interrupt channel */
    VOIDFUNCPTR proutine   /* routine to call on int */
)
{
    long status;

    if (prm == NULL)
       return S_dev_NoInit;
    if (irqNumber < 0 || irqNumber > 3)
       return S_dev_vxWorksVecInstlFail;
    if (pisr[irqNumber] != NULL)
       return S_dev_vectorInUse;

    /* plug in our wrapper routine */
    status = intConnect((void *) (RM_INT_VECTOR + irqNumber), 
             vmi5588_intr, irqNumber);
    if (status != OK)
       return status;

    /* save the routine pointer */
    pisr[irqNumber] = proutine;

    /* clear out card interrupt FIFOs */
    if (irqNumber > 0)
       prm->interrupt[irqNumber].senderId = 0;

    /* finally enable the hardware */
    prm->interrupt[irqNumber].control = RM_INT_LEVEL | 
                                        RM_CR_INT_ENABLE | 
                                        RM_CR_INT_AUTOCLR;
    return OK;
}

/*****************************************************************************
*
* rmIntDisconnect - Disconnect an RM Interrupt routine
*
* This disables RM interrupts on the given channel number and marks the
* channel as unused.
*
* RETURNS:
* OK, or S_dev_???
*
* EXAMPLE:
* .CS
*   int status;
*
*   status = rmIntDisconnect(2);
*   if (status)
*       printf("No ISR Connected\\n");
* .CE
*
* SEE ALSO: rmIntConnect()
*/

long rmIntDisconnect
(
    int irqNumber      /* RM interrupt channel */
)
{
    if (prm == NULL)
       return S_dev_NoInit;
    if (irqNumber < 0 || irqNumber > 3 || pisr[irqNumber] == NULL)
       return S_dev_vectorNotInUse;

    /* disable the hardware */
    prm->interrupt[irqNumber].control &= ~RM_CR_INT_ENABLE;

    /* flag interrupt unused */
    pisr[irqNumber] = NULL;

    return OK;
}

/*****************************************************************************
* rmIntSend - send an RM interrupt
*
* This causes an interrupt to be sent out on the RM bus, using the given
* channel number.  The interrupt will be broadcast to all other RM nodes
* if the node ID given is -1, otherwise it will be sent only to the node
* specified.
*
* You can't send an interrupt to yourself.  The hardware will ignore it.
*
* RETURNS:
* OK, or S_dev_???
*
* EXAMPLE:
* .CS
*   #define BROADCAST -1
*
*   rmIntSend(2, BROADCAST);
* .CE
*
* SEE ALSO: rmIntConnect(), rmIntDisconnect()
*/

long rmIntSend
(
    int irqNumber,     /* RM interrupt channel */
    int nodeId         /* node number, or -1 for broadcast */
)
{
    if (prm == NULL)
       return S_dev_NoInit;
    if (irqNumber < 1 || irqNumber > 3 || nodeId == prm->nodeId)
       return S_dev_badRequest;

    if (nodeId == -1)
       /* Broadcast to all nodes */
       prm->cmd_Node = irqNumber << 8 | RM_CMD_BROADCAST;
    else
       /* Node specific */
       prm->cmd_Node = irqNumber << 8 | (nodeId & 0xff);

    return OK;
}

/*****************************************************************************
*
* rmNodeId - return my RM Node number
*
* Used to find the value of the RM node ID setting on the VMIC5588 card.
*
* RETURNS:
* vmi5588 Node number (in the range 0 to 255), or S_dev_NoInit.
*
* EXAMPLE:
* .CS
*   long me;
*
*   me = rmNodeId();
*   if (me > 255L)
*       printf("RM error %d\\n", me);
* .CE
*
*/

long rmNodeId
(
    void
)
{
    if (prm == NULL)
       return S_dev_NoInit;

    return prm->nodeId;
}


/*****************************************************************************
*
* rmStatus - return current RM status information
*
* This routine reads various  status bits from the 5588 card and returns a
* composite of the interesting bits.  The input parameter is a bit mask
* which is used to reset those status bits which are writable -- if rmStatus
* is called with the parameters RM_RESYNC, RM_BADXFER or RM_NORING then the
* relevant status bit or bits will be cleared if they are currently set.  
* To clear all status bits, use a parameter value of 0xffff.  The returned
* status value is not affected by the input parameter.
*
* The RM_NORING bit functions slightly differently to the others. Whenever 
* rmStatus is called with the the bit RM_NORING set the current value of
* the status bit is returned and the bit cleared as normal, but the test
* location in the reflected memory is also written which results in a data
* packet being transmitted around the ring.  Any calls to rmStatus which 
* occur before this packet (or any other packets which originated at this
* node) has been received will return RM_NORING, but the ring transition 
* time should only be a few microseconds at worst.
*
* RETURNS:
* Status from the VMIC5588 card (a value in the range 0 to 0xffff), or 
* S_dev_NoInit.
*
* The following symbols are defined in the header file vmi5588.h and 
* provide a bit mask for the return state.  The bit will be set if the 
* condition described below is true, so zero means Good status.
*
* .CS
*    RM_IRQ1     Interrupt 1 Pending
*    RM_IRQ2     Interrupt 2 Pending
*    RM_IRQ3     Interrupt 3 Pending
*    RM_NOSIG    No Input Signal
*    RM_NOSYNC   Input PLL unsynchronized
*    RM_RESYNC   PLL Recently unsynchronized
*    RM_NORING   Fibre Ring broken
*    RM_BADXFR   Single Transfer Error
*    RM_TXHALF   Transmit FIFO Half-full
*    RM_RXHALF   Receive FIFO Half-full
* .CE
*
* EXAMPLE:
* .CS
*   -> rmStatus 0x40
*   value = 64 = 0x44 = '@'
*   -> rmStatus
*   value = 4 = 0x4
* .CE
*
*/

long rmStatus
(
    long reset      /* Status bits to reset */
)
{
    long            stat;
    static unsigned char testCounter;

    if (prm == NULL)
       return S_dev_NoInit;

    /*
     * The following combines the two status registers, masks out the bits
     * we're not interested in and makes everything active high, so 0=good
     * status. Note 0x40 = RM_RESYNC is probably good status, but indicates
     * that we lost sync recently, ie someone got turned off & on again.
     */
    stat = ((prm->intRxStatus | prm->boardCsr << 8) &
           (RM_RESYNC | RM_NOSYNC | RM_NOSIG  | RM_IRQ3 | RM_IRQ2 | RM_IRQ1 |
            RM_RXHALF | RM_TXHALF | RM_BADXFR | RM_NORING)) ^
           (RM_TXHALF | RM_RXHALF | RM_NORING);

    /* Clear the R/W status bits */
    if (stat & reset & RM_RESYNC)
       prm->intRxStatus &= ~RM_IRS_LATCHED;
    if (stat & reset & RM_BADXFR)
       prm->boardCsr &= ~RM_CSR_BAD_DATA;
    if (reset & RM_NORING) {
       prm->boardCsr &= ~RM_CSR_OWN_DATA;
       prm->test = (prm->nodeId << 8) + testCounter++;
    }

    return stat;
}

/*****************************************************************************
*
* rmLoadSymbols - load RM symbol file from stdin
*
* This routine expects to have a file connected to its stdin, and reads
* a reflective memory allocation map from that file. For EPICS to start up
* properly, the symbols should be loaded before iocInit is called so that
* the RM records can find their hardware addresses at initialisation, thus
* it is not necessary for vmi5588_init to have been called before loading
* the symbol file.
*
* The file format comprises a series of lines, each starting with a keyword,
* possible followed by a name, and for some keywords a number is also used.
* Comment lines should have a # symbol in the first column, and blank lines
* may also be included. Each keyword describes and optionally names a record
* to appear in the reflected memory map, or declares the start of a new RM
* page. The syntax is as follows:
*
* .CS
*          page [name [nnn]]
*          analogue [name]
*          long [name]
*          string [name]
*          array name nnn
*          user name nnn
* .CE
*
* If no name is given for a particular record, a gap will be left in the
* memory map of the correct size but the record cannot be accessed from this
* IOC. This permits different files to be used for different IOCs, naming 
* only those records to be used but leaving space for the other records. The
* page structure can be used as an alternative to this, and it is possible
* to load several different symbol files one after the other, having a single
* page declaration in each say.
*
* The optional number in a page declaration makes it possible to declare RM
* pages in any order. Note that no checking is performed to see if a page has 
* already been declared, so it is possible to have duplicate and conflicting
* declarations. The page name is created as a symbol, but cannot be used to
* look up an address.
*
* An array record can contain data of any type (set by FTVL field) and 
* number of elements (set by NELM field), so a name storage size (in bytes)
* must be given in the symbol file.  To calculate the amount of storage
* needed, multiply the number of elements in the array by the size of a
* single element according to the following table.  An overhead of 16 bytes
* is automatically added to the figure given and the length rounded up to a
* 4-byte boundary when actually allocating the storage.
*
* .CS
*          Field Type    Size
*          ----------    ----
*          STRING         40
*          CHAR/UCHAR      1
*          SHORT/USHORT    2
*          LONG/ULONG      4
*          FLOAT           4
*          DOUBLE          8
*          ENUM            2
* .CE
*
* The user keyword allows areas of RM to be defined for fast access via the C
* subroutine interface. For this keyword, both the name and size parameter
* (which is the number of bytes of storage desired) are required.
*
* RETURNS:
* OK, or S_dev_???
*
* EXAMPLE:
* .CS
*   -> rmLoadSymbols < synchro/mysym.rms
*   value = 0 = 0x0
*   -> rmLoadSymbols
*   page seven 7
*   analogue Command
*   # array of 5 * DOUBLE = (5*8)+16 bytes
*   array Arguments 56
*   page test 200
*   user shared 64
*   ^Dvalue = 0 = 0x0
*
* SEE ALSO: rmPrintSymbols(), rmAddr()
*/

long rmLoadSymbols
(
    void
)
{
    int             pageNumber = 0,
                    lineNumber = 0,
                    pageOffset = 0;
    int             tokenCount, i;
    long            status;
    char            inputLine[RM_INPUTLINESIZE];
    char            inputType[RM_INPUTTYPESIZE];
    char            inputName[RM_INPUTNAMESIZE];
    short           inputNumber;

    /* create symbol table if not defined */
    if (rmSymTbl == NULL) {
       rmSymTbl = symTblCreate(RM_SYM_HASHSIZE, FALSE, memSysPartId);
       if (rmSymTbl == NULL) {
          fprintf(stderr, "rmLoadSymbols: Can't create RM symbol table\n");
          return S_dev_noMemory;
       }
    }

    status = OK;

    /* scan file until EOF */
    while (gets(inputLine)) {
      lineNumber++;
   
      /* ignore if it's a blank line or comment */
      if (strlen(inputLine) == 0)
         continue;
      if (inputLine[0] == '#')
         continue;

      /* split line into <type> [<name> [<number>]] */
      tokenCount = sscanf(inputLine, "%s%s%hi",
                   inputType, inputName, &inputNumber);

      if (tokenCount < 1)
          continue;

      /* look up this type */
      for (i = 0; rmToken[i].ptokString != NULL; i++)
          if (strcmp(inputType, rmToken[i].ptokString) == 0)
             break;

      switch (rmToken[i].tokType) {
         case RM_TYPE_PAGE:   /* page [<name> [<number>]] */

              /* start a new rm page */
              pageNumber++;
              pageOffset = 0;
              if (tokenCount > 2)
                 if (inputNumber < 0 || inputNumber > RM_NUM_PAGE) {
                    status = S_dev_badSignalNumber;
                    fprintf(stderr, 
                     "rmLoadSymbols: Bad RM page number at line %d (name %s)\n",
                     lineNumber, inputName);
                 }
                 else
                    pageNumber = inputNumber;
              if (tokenCount > 1 && 
                  symAdd(rmSymTbl, inputName, 
                         (char *)(pageNumber * RM_PAGE_SIZE + pageOffset), 
                         (SYM_TYPE)RM_TYPE_PAGE, 0)) {
                 status = S_dev_multDevice;
                 fprintf(stderr,
                   "rmLoadSymbols: Duplicate RM page name at line %d: %s\n",
                   lineNumber, inputName);
              }
              break;

         case RM_TYPE_ALOG:   /* analogue [<name>] */
              if (tokenCount > 1 && 
                  symAdd(rmSymTbl, inputName,
                         (char *)(pageNumber * RM_PAGE_SIZE + pageOffset),
                         RM_TYPE_ALOG, 0)) {
                 status = S_dev_multDevice;
                 fprintf(stderr,
                 "rmLoadSymbols: Duplicate RM analogue symbol at line %d: %s\n",
                 lineNumber, inputName);
              }
              pageOffset += RM_SIZE_ALOG;
              break;

         case RM_TYPE_LONG:   /* long [<name>] */
              if (tokenCount > 1 && 
                  symAdd(rmSymTbl, inputName,
                         (char *) (pageNumber * RM_PAGE_SIZE + pageOffset),
                         RM_TYPE_LONG, 0)) {
                 status = S_dev_multDevice;
                 fprintf(stderr,
                    "rmLoadSymbols: Duplicate RM long symbol at line %d: %s\n",
                    lineNumber, inputName);
              }
              pageOffset += RM_SIZE_LONG;
              break;

         case RM_TYPE_STRG:   /* string [<name>] */
              if (tokenCount > 1 && 
                  symAdd(rmSymTbl, inputName,
                         (char *) (pageNumber * RM_PAGE_SIZE + pageOffset),
                         RM_TYPE_STRG, 0)) {
              status = S_dev_multDevice;
              fprintf(stderr,
                 "rmLoadSymbols: Duplicate RM string symbol at line %d: %s\n",
                 lineNumber, inputName);
              }
              pageOffset += RM_SIZE_STRG;
              break;

         case RM_TYPE_USER:   /* user <name> <size> */
              if (tokenCount < 3) {
                 status = S_dev_badRequest;
                 fprintf(stderr, 
                    "rmLoadSymbols: Missing user parameter(s) at line %d\n", 
                    lineNumber);
              }
              else {
                 if (symAdd(rmSymTbl, inputName,
                            (char *)(pageNumber * RM_PAGE_SIZE + pageOffset),
                            RM_TYPE_USER, 0)) {
                    status = S_dev_multDevice;
                    fprintf(stderr,
                     "rmLoadSymbols: Duplicate RM user symbol at line %d: %s\n",
                     lineNumber, inputName);
                  }
                  pageOffset += 
                  (inputNumber & 3) ? (inputNumber & ~3) + 4 : inputNumber;
              }
              break;

         case RM_TYPE_ARRY:   /* array <name> <size> */
              if (tokenCount < 3) {
                 status = S_dev_badRequest;
                 fprintf(stderr, 
                    "rmLoadSymbols: Missing array parameter(s) at line %d\n", 
                    lineNumber);
              }
              else {
                 if (symAdd(rmSymTbl, inputName,
                            (char *)(pageNumber * RM_PAGE_SIZE + pageOffset),
                            RM_TYPE_ARRY, 0)) {
                    status = S_dev_multDevice;
                    fprintf(stderr,
                    "rmLoadSymbols: Duplicate RM array symbol at line %d: %s\n",
                    lineNumber, inputName);
                 }
                 inputNumber += RM_SIZE_ARRY;
                 pageOffset += 
                 (inputNumber & 3) ? (inputNumber & ~3)+4 : inputNumber;
              }
              break;

         default:      /* unrecognised */
              status = S_dev_badRequest;
              fprintf(stderr,
                      "rmLoadSymbols: Unrecognised RM type at line %d: %s\n", 
                      lineNumber, inputType);
      } /* switch */

      if (pageOffset > RM_PAGE_SIZE) {
          status = S_dev_noMemory;
          fprintf(stderr,
             "rmLoadSymbols: RM Page overflow at line %d, page %d\n", 
             lineNumber, pageNumber);
          /* Try and recover */
          pageNumber += (pageOffset / RM_PAGE_SIZE);
          pageOffset = pageOffset % RM_PAGE_SIZE;
      }
    } /* while */

    clearerr(stdin);

    return status;
}

/*****************************************************************************
*
* rmAddr - lookup address in RM symbol table
*
* This routine looks up the symbol with given name and type, and returns
* the address of the VMIC5588 storage for that record.
*
* Symbol types are defined by a series of definitions in the vmi5588.h
* header file.  All C subroutine interfaces should use RM_TYPE_USER
* for their memory areas.
*
* .CS
*          Type Name     Value
*          ---------     -----
*          RM_TYPE_ALOG   0x01
*          RM_TYPE_LONG   0x02
*          RM_TYPE_STRG   0x03
*          RM_TYPE_ARRY   0x04
*          RM_TYPE_USER   0x10
*          RM_TYPE_PAGE   0x80
* .CE
*
* RETURNS
* RM address, or NULL if no such symbol or driver not initialised.
*
* EXAMPLE
* .CS
*   double *pshared;
*
*   pshared = rmAddr("Zernike", RM_TYPE_USER);
*   if (pshared == NULL) {
*       printf("Can't locate Zernike area in RM\\n");
*       return(ERROR);
*   }
* .CE
* 
* SEE ALSO: rmLoadSymbols(), rmPrintSymbols()
*/

void *rmAddr
(
    char *pname,      /* symbol name to find */
    int rmType         /* RM_TYPE_??? */
)
{
    int             offset;
    SYM_TYPE        retType;
    STATUS          status;

    if (prm == NULL)
       return NULL;

    status = symFindByNameAndType(rmSymTbl, pname, (char **) &offset, &retType,
                                  (SYM_TYPE) rmType, (SYM_TYPE) ~0);

    return status ? NULL : &prm->mem[offset];
}


/*****************************************************************************
*
* rmLookup - lookup Address, Page and Offset of RM symbol
*
* This routine looks up the symbol with given name and type, and returns
* the associated page and offset values.
*
* RETURNS:
* OK, or S_dev_??? if no such symbol or driver not initialised.
*
* NOMANUAL
*/

long rmLookup
(
    char *pname,                 /* symbol name to find */
    int rmType,                  /* RM_TYPE_??? */
    struct rm_data **pprmData,   /* symbol address location */
    short *prmPage,              /* symbol page location */
    short *prmOffset             /* symbol offset location */
)
{
    int             offset;
    SYM_TYPE        retType;

    if (prm == NULL)
       return S_dev_NoInit;

    if (symFindByNameAndType(rmSymTbl, pname, (char **) &offset, &retType,
                             (SYM_TYPE) rmType, (SYM_TYPE) ~0))
       return S_dev_badSignal;

    *prmPage = offset / RM_PAGE_SIZE;
    *prmOffset = offset % RM_PAGE_SIZE;
    *pprmData = (struct rm_data *) &prm->mem[offset];
    return OK;
}

/*****************************************************************************
*
* printSym - print symbol callback routine.
*
* Prints name string, address (offset if prm not set) and type. Callback
* routine for use by symEach() from rmPrintSymbols().
*
* RETURNS:
* TRUE, so symEach() continues
*
* NOMANUAL
*/

LOCAL BOOL printSym
(
    char *pname,
    int offset,
    SYM_TYPE type,
    int arg,
    UINT16 group
)
{
    int i;

    for (i=0; rmToken[i].ptokString != NULL; i++)
        if ((type & 0xff) == rmToken[i].tokType)
           break;

    printf("%-15.15s  %8X   %s\n", pname,
           (prm == NULL) ? offset : (unsigned int) &prm->mem[offset], 
           rmToken[i].ptokString);
    return TRUE;
}


/*****************************************************************************
*
* rmPrintSymbols - display all RM symbols
*
* This command prints the name, address and type of all known RM symbols.
* If the driver has not been initialised yet, the offset into the 5588
* shared memory area is displayed instead of the address.
*
* EXAMPLE:
* .CS
*   -> rmPrintSymbols
*   Name               Addr     Type
*   A3               F0200730   analogue
*   S7               F0201B90   string
*   TrigL0           F0200F00   long
*   S1               F0201730   string
* .CE
*
* SEE ALSO: rmLoadSymbols(), rmAddr()
*/

void rmPrintSymbols
(
    void
)
{
    printf("Name               Addr     Type\n");
    symEach(rmSymTbl, (FUNCPTR) printSym, 0);
}

#ifndef NO_EPICS
/*****************************************************************************
*
* vmi5588_pageISR - page trigger interrupt service routine
*
* Called by an incoming I/O Interrupt on channel 1, causes records
* in modified pages which are set to I/O Interrupt scanning to be
* processed.
*
* RETURNS: N/A
*
* NOMANUAL
*/

void vmi5588_pageISR
(
    int rmNodeId       /* source Node Id */
)
{
    short i;

    if (vmi5588Debug) {
        logMsg("vmi5588_pageISR: Interrupt from Node <%x>\n", 
               rmNodeId, 0,0,0,0,0);
    }

    /* check for new data in page */
    for (i = 0; i < pageIo.number; i++)
        if (prm->pageFlag[pageIo.p[i].page] != pageIo.p[i].lastPageFlag) {
           pageIo.p[i].lastPageFlag = prm->pageFlag[pageIo.p[i].page];

           if (vmi5588Debug) {
              logMsg("Triggering page <%d>\n", pageIo.p[i].page, 0,0,0,0,0);
           }

           scanIoRequest(pageIo.p[i].ioscanpvt);
        }
}
#endif

#ifndef NO_EPICS
/*****************************************************************************
*
* vmi5588_pageInit - initialise page data structure
*
* Initialises the page structure used for incoming I/O Interrupts.
* No action if page has already been initialised.
*
* RETURNS:
* OK, or S_dev_??? if can't connect interrupt
*
* NOMANUAL
*/

long vmi5588_pageInit
(
    short rmPage       /* rm page number */
)
{
    short pageIndex;
    long status = OK;

    for (pageIndex = 0; pageIndex < pageIo.number; pageIndex++)
        if (rmPage == pageIo.p[pageIndex].page)
           return OK;

    if (pageIndex == 0)
       if ((status = rmIntConnect(1, vmi5588_pageISR)) != OK)
          fprintf(stderr, "vmi5588_pageInit: Can't connect RM irq #1\n");

    pageIo.p[pageIndex].page = rmPage;
    scanIoInit(&pageIo.p[pageIndex].ioscanpvt);
    pageIo.number++;

    if (vmi5588Debug) {
       printf("Init pageIo index <%hd> page <%d>\n",
              pageIndex, (int)pageIo.number);
    }

    return status;
}
#endif

#ifndef NO_EPICS
/*****************************************************************************
*
* vmi5588_pvtInit - initialise private data structure
*
* Allocates storage for and initialises private record data, returning
* a pointer to the initialised structure in the first parameter.
*
* RETURNS: N/A
*
* NOMANUAL
*/

void vmi5588_pvtInit
(
    struct rmpvt **ppdpvt,    /* location for private pointer */
    short rmPage,             /* rm page number */
    short rmOffset,           /* offset into rm page */
    struct rm_data *prmData   /* pointer to shared memory */
)
{
   struct rmpvt   *prmpvt;

   prmpvt = (struct rmpvt *) calloc(1, sizeof(struct rmpvt));
   prmpvt->page = rmPage;
   prmpvt->offset = rmOffset;
   prmpvt->address = prmData;

#ifdef RM_DEBUG
   printf("page<%x> offset<%x> address<%x>\n", 
          rmPage, rmOffset, (unsigned long) prmData);
#endif

   *ppdpvt = prmpvt;
}
#endif

#ifndef NO_EPICS
/*****************************************************************************
*
* vmi5588_getIoscanpvt - look up I/O interrupt data structure
*
* Returns the IOSCANPVT token for a record given its dpvt structure.
*
* RETURNS: OK, or S_dev_internal if no entry found.
*
* NOMANUAL
*/

long vmi5588_getIoscanpvt
(
    struct rmpvt *pdpvt,   /* record private data structure */
    IOSCANPVT *pscanpvt    /* location for interrupt data */
)
{
    short i, page;

    page = pdpvt->page;
    for (i = 0; i < pageIo.number; i++)
    if (page == pageIo.p[i].page) {
       *pscanpvt = pageIo.p[i].ioscanpvt;
       return OK;
    }

/* Question: could I call vmi5588_pageInit at this point, since it
   obviously hasn't been done yet?  This would allow a record to be
   set to I/O Interrupt Scanning after initialisation.
   It may be illegal to call ScanIoInit now.

   Answer: It's legal; add this when I get the chance.
*/

    return S_dev_internal;
}
#endif

#ifndef NO_EPICS
/*****************************************************************************
*
* vmi5588_trigger - trigger a remote I/O interrupt
*
* Sends a broadcast interrupt to trigger I/O interrupt processing for all
* RM pages on given page number.
*
* RETURNS: OK, or S_dev_???
*
* NOMANUAL
*/

long vmi5588_trigger
(
    short rmPage      /* rm page number */
)
{
    static unsigned char nodeIDcounter;

    prm->pageFlag[rmPage] = (prm->nodeId << 8) + nodeIDcounter++;

    if (vmi5588Debug) {
       logMsg("vmi5588_trigger: Broadcast trigger for page <%hd>\n", 
              rmPage, 0,0,0,0,0);
    }

    return rmIntSend(1, -1);
}
#endif

